#include "stdafx.h"
#include "../ist/ist.h"
#include "../types.h"
#include "Message.h"
#include "Fraction.h"
#include "FractionTask.h"
#include "FractionCollider.h"

namespace atomic
{



Task_FractionUpdate::Task_FractionUpdate()
{
    m_sortx_task = EA_NEW(Task_FractionSortX) Task_FractionSortX();
    m_sorty_task = EA_NEW(Task_FractionSortY) Task_FractionSortY();
    m_sortz_task = EA_NEW(Task_FractionSortZ) Task_FractionSortZ();
}

Task_FractionUpdate::~Task_FractionUpdate()
{
    for(size_t i=0; i<m_move_tasks.size(); ++i) {
        EA_DELETE(m_move_tasks[i]);
    }
    for(size_t i=0; i<m_col_test_tasks.size(); ++i) {
        EA_DELETE(m_col_test_tasks[i]);
    }
    for(size_t i=0; i<m_col_proc_tasks.size(); ++i) {
        EA_DELETE(m_col_proc_tasks[i]);
    }
    m_move_tasks.clear();
    m_col_test_tasks.clear();
    m_col_proc_tasks.clear();
    EA_DELETE(m_sortx_task);
    EA_DELETE(m_sorty_task);
    EA_DELETE(m_sortz_task);
}

void Task_FractionUpdate::initialize(FractionSet *obj)
{
    m_obj = obj;
}

void Task_FractionUpdate::exec()
{
    TaskScheduler* scheduler = TaskScheduler::getInstance();

    // bZ[W
    m_obj->processGenerateMessage();

    uint32 num_blocks = m_obj->getNumBlocks();
    m_blocks = num_blocks;
    // Փˊƃ^XNubNTCYɍ킹
    FractionSet::getInterframe()->resizeColliders(num_blocks);
    while(m_move_tasks.size()<num_blocks) {
        m_move_tasks.push_back(EA_NEW(Task_FractionMove) Task_FractionMove());
    }
    while(m_col_test_tasks.size()<num_blocks) {
        m_col_test_tasks.push_back(EA_NEW(Task_FractionCollisionTest) Task_FractionCollisionTest());
    }
    while(m_col_proc_tasks.size()<num_blocks) {
        m_col_proc_tasks.push_back(EA_NEW(Task_FractionCollisionProcess) Task_FractionCollisionProcess());
    }

    // ړ^XNXPW[&s҂
    for(uint32 i=0; i<num_blocks; ++i) {
        m_move_tasks[i]->initialize(m_obj, i);
    }
    scheduler->schedule((Task**)&m_move_tasks[0], num_blocks);
    scheduler->waitFor((Task**)&m_move_tasks[0], num_blocks);

    // \[g^XNXPW[&s҂
    m_sortx_task->initialize(m_obj);
    m_sorty_task->initialize(m_obj);
    m_sortz_task->initialize(m_obj);
    Task *sorts[] ={m_sortx_task, m_sorty_task, m_sortz_task};
    scheduler->schedule(sorts, _countof(sorts));
    scheduler->waitFor(sorts, _countof(sorts));

    // Փ˔^XNXPW[&s҂
    for(uint32 i=0; i<num_blocks; ++i) {
        m_col_test_tasks[i]->initialize(m_obj, i);
    }
    scheduler->schedule((Task**)&m_col_test_tasks[0], num_blocks);
    scheduler->waitFor((Task**)&m_col_test_tasks[0], num_blocks);

    // Փːis^XNXPW[&s҂
    for(uint32 i=0; i<num_blocks; ++i) {
        m_col_proc_tasks[i]->initialize(m_obj, i);
    }
    scheduler->schedule((Task**)&m_col_proc_tasks[0], num_blocks);
    scheduler->waitFor((Task**)&m_col_proc_tasks[0], num_blocks);

}


void Task_FractionUpdate::waitForCompletion()
{
    uint32 num_blocks = m_blocks;
    TaskScheduler* scheduler = TaskScheduler::getInstance();
    scheduler->waitFor((Task**)&m_move_tasks[0], num_blocks);
    scheduler->waitFor((Task**)&m_col_test_tasks[0], num_blocks);
    scheduler->waitFor((Task**)&m_col_proc_tasks[0], num_blocks);

    Task *sorts[] ={m_sortx_task, m_sorty_task, m_sortz_task};
    scheduler->waitFor(sorts, _countof(sorts));
}


} // namespace atomic
